from enum import Enum
from typing import Any, Generator, Iterable

from nomenklatura.dataset import DataCatalog, Dataset


def StrEnum(name: str, values: Iterable[Any]) -> Enum:
    # mimic py3.11 enum.StrEnum
    class _StrEnum(str, Enum):
        def __str__(self):
            return self.value

    return _StrEnum(name, {str(v): str(v) for v in values})


def make_dataset(name: str) -> Dataset:
    catalog = DataCatalog(
        Dataset, {"datasets": [{"name": name, "title": name.title()}]}
    )
    return catalog.get(name)


def parse_unknown_cli_filters(
    filters: tuple[str],
) -> Generator[tuple[str, str, str], None, None]:
    filters = (f for f in filters)
    for prop in filters:
        prop = prop.lstrip("-")
        if "=" in prop:  # 'country=de'
            prop, value = prop.split("=")
        else:  # ("country", "de"):
            value = next(filters)

        prop, *op = prop.split("__")
        yield prop, value, op[0] if op else None



from typing import Any, Generator, Literal, TypeAlias

from nomenklatura.entity import CE

from .enums import Properties, Schemata

# a string-keyed dict
SDict: TypeAlias = dict[str, Any]

# property multi-value
Value: TypeAlias = list[str]

# composite entity generator
CEGenerator: TypeAlias = Generator[CE, None, None]

StrGenerator: TypeAlias = Generator[str, None, None]
BytesGenerator: TypeAlias = Generator[bytes, None, None]

Schemata: TypeAlias = Literal[tuple(s.name for s in Schemata)]
Properties: TypeAlias = Literal[tuple(p.name for p in Properties)]



import contextlib
import sys
from typing import Any, Iterable, Literal

import orjson
from banal import is_listish
from followthemoney import model
from nomenklatura.entity import CE, CompositeEntity
from nomenklatura.util import PathLike
from smart_open import open

from .types import CEGenerator, SDict


def load_proxy(data: dict[str, Any]) -> CE:
    return CompositeEntity.from_dict(model, data)


@contextlib.contextmanager
def smart_open(
    uri: str | None = None,
    sys_io: Literal[sys.stdin.buffer, sys.stdout.buffer] | None = sys.stdin,
    **kwargs
):
    if uri and uri != "-":
        fh = open(uri, **kwargs)
    else:
        fh = sys_io

    try:
        yield fh
    finally:
        if fh not in (sys.stdout.buffer, sys.stdin.buffer):
            fh.close()


def smart_read_proxies(
    uri: PathLike | Iterable[PathLike],
    mode: str | None = "rb",
    serialize: bool | None = True,
) -> CEGenerator:
    if is_listish(uri):
        for u in uri:
            yield from smart_read_proxies(u, mode, serialize)
        return

    with smart_open(uri, sys.stdin.buffer, mode=mode) as fh:
        while True:
            line = fh.readline()
            if not line:
                break
            data = orjson.loads(line)
            if serialize:
                data = load_proxy(data)
                data.datasets.discard("default")
            yield data


def smart_write_proxies(
    uri: PathLike,
    proxies: Iterable[CE | SDict],
    mode: str | None = "wb",
    serialize: bool | None = False,
) -> int:
    ix = 0
    with smart_open(uri, sys.stdout.buffer, mode=mode) as fh:
        for proxy in proxies:
            ix += 1
            if serialize:
                proxy = proxy.to_dict()
            fh.write(orjson.dumps(proxy, option=orjson.OPT_APPEND_NEWLINE))
    return ix


def apply_datasets(
    proxies: Iterable[CE], *datasets: Iterable[str], replace: bool | None = False
) -> CEGenerator:
    for proxy in proxies:
        if datasets:
            if replace:
                proxy.datasets = set(datasets)
            else:
                proxy.datasets.update(datasets)
        yield proxy



from typing import Iterable, Optional, TypedDict, TypeVar

from nomenklatura.entity import CE

from .filters import (
    Dataset,
    DatasetFilter,
    F,
    Property,
    PropertyFilter,
    Schema,
    SchemaFilter,
    Value,
)
from .types import CEGenerator

Q = TypeVar("Q", bound="Query")
L = TypeVar("L", bound="Lookup")


class Lookup(TypedDict):
    dataset: Optional[Dataset | str]
    schema: Optional[Schema | str]
    prop: Optional[Property | str]
    value: Optional[Value]


class Query:
    filters: set[F] = set()

    def __init__(self, *filters: Iterable[F]):
        self.filters = {f for f in filters}

    def where(self, **lookup: Lookup) -> Q:
        if "dataset" in lookup:
            self.filters.add(DatasetFilter(lookup["dataset"]))
        if "schema" in lookup:
            self.filters.add(
                SchemaFilter(
                    lookup["schema"],
                    include_descendants=lookup.pop("include_descendants", False),
                    include_matchable=lookup.pop("include_matchable", False),
                )
            )
        if "prop" in lookup and "value" in lookup:
            self.filters.add(
                PropertyFilter(
                    lookup["prop"], lookup["value"], lookup.pop("operator", None)
                )
            )
        return self.__class__(*self.filters)

    def apply(self, proxy: CE) -> bool:
        if not self.filters:
            return True
        return all(f.apply(proxy) for f in self.filters)

    def apply_iter(self, proxies: CEGenerator) -> CEGenerator:
        """
        apply a `Query` to a generator of proxies and return a generator of filtered proxies
        """
        for proxy in proxies:
            if self.apply(proxy):
                yield proxy



from typing import TypeVar

from banal import as_bool, ensure_list
from followthemoney import model
from followthemoney.property import Property
from followthemoney.schema import Schema
from nomenklatura.dataset import Dataset
from nomenklatura.entity import CE

from .enums import Operators, Properties, Schemata
from .exceptions import ValidationError
from .types import Value
from .util import StrEnum, make_dataset


class BaseFilter:
    instance: Dataset | Schema | Property | None = None
    options: StrEnum = None

    def __init__(self, value: str | Dataset | Schema | Property | None):
        self.instance = self.get_instance(value)
        self.validate()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: `{self.instance}`>"

    def __str__(self) -> str:
        return str(self.instance)

    def validate(self):
        if self.options is None:
            return
        try:
            self.options[self.get_str_value()]
        except (KeyError, AttributeError):
            raise ValidationError(f"{self}: invalid value: `{self.instance}`")

    def apply(self, proxy: CE) -> bool:
        raise NotImplementedError

    def get_instance(self, value: str) -> Dataset | Schema | Property:
        raise NotImplementedError

    def get_str_value(self) -> str:
        return str(self.instance)


class DatasetFilter(BaseFilter):
    instance: Dataset = None

    def apply(self, proxy: CE) -> bool:
        return self.instance.name in proxy.datasets

    def get_instance(self, value: str | Dataset) -> Dataset:
        if isinstance(value, str):
            value = make_dataset(value)
        return value


class SchemaFilter(BaseFilter):
    instance: Schema = None
    options = Schemata

    def __init__(
        self,
        schema: str | Schema,
        include_descendants: bool = False,
        include_matchable: bool = False,
    ):
        super().__init__(schema)
        self.values: set[Schema] = {self.instance}
        self.include_descendants = include_descendants
        self.include_matchable = include_matchable
        if self.include_descendants:
            self.values.update(self.instance.descendants)
        if self.include_matchable:
            self.values.update(self.instance.matchable_schemata)

    def apply(self, proxy: CE) -> bool:
        for schema in self.values:
            if proxy.schema.name == schema.name:
                return True
        return False

    def get_instance(self, value: str | Schema) -> Schema:
        if isinstance(value, str):
            value = model.get(value)
        return value

    def get_str_value(self) -> str:
        return self.instance.name


class Operator:
    def __init__(self, operator: Operators, value: str | None = None):
        self.operator = self.get_operator(operator)
        self.value = value

    def get_value(self):
        if self.operator == "in":
            return ensure_list(self.value)
        if self.operator == "null":
            return as_bool(self.value)
        return str(self.value) if self.value is not None else None

    def get_operator(self, operator: str) -> Operators:
        try:
            return Operators[operator]
        except KeyError:
            raise ValidationError(f"{self}: Invalid oparator: `{operator}`")

    def apply(self, value: str | None) -> bool:
        parsed_value = self.get_value()
        if self.operator == "in":
            return value in parsed_value
        if self.operator == "null":
            return not value == parsed_value
        if self.operator == "gt":
            return value > parsed_value
        if self.operator == "gte":
            return value >= parsed_value
        if self.operator == "lt":
            return value < parsed_value
        if self.operator == "lte":
            return value <= parsed_value


class PropertyFilter(BaseFilter):
    instance: Property = None
    options = Properties
    value: Value = None
    operator: Operator = None

    def __init__(self, prop: Property, value: Value, operator: str | None = None):
        super().__init__(prop)
        self.value = value
        if operator is not None:
            self.operator = Operator(operator, value)
        else:
            self.operator = None

    def apply(self, proxy: CE) -> bool:
        values = proxy.get(self.instance.name, quiet=True)
        if self.operator is not None:
            for value in values:
                if self.operator.apply(value):
                    return True
        else:
            return self.value in values

    def get_instance(self, value: str | Property) -> Property:
        if isinstance(value, str):
            for prop in model.properties:
                if prop.name == value or prop.qname == value:
                    return prop
        return value

    def get_str_value(self):
        return self.instance.name


Filter = DatasetFilter | SchemaFilter | PropertyFilter
F = TypeVar("F", bound=Filter)



from followthemoney import model

from .util import StrEnum

Schemata = StrEnum("Schemata", [k for k in model.schemata.keys()])
Properties = StrEnum("Properties", [n for n in {p.name for p in model.properties}])
Operators = StrEnum("Operators", ["in", "null", "gt", "gte", "lt", "lte"])



import click
from click_default_group import DefaultGroup

from ftmq.io import apply_datasets, smart_read_proxies, smart_write_proxies

from .query import Query
from .util import parse_unknown_cli_filters


@click.group(cls=DefaultGroup, default="q", default_if_no_args=True)
def cli():
    pass


@cli.command(
    context_settings=dict(
        ignore_unknown_options=True,
    ),
)
@click.option(
    "-i", "--input-uri", default="-", show_default=True, help="input file or uri"
)
@click.option(
    "-o", "--output-uri", default="-", show_default=True, help="output file or uri"
)
@click.option("-d", "--dataset", multiple=True, help="Dataset(s) to filter for")
@click.option("-s", "--schema", multiple=True, help="Schema(s) to filter for")
@click.option(
    "--schema-include-descendants", is_flag=True, default=False, show_default=True
)
@click.option(
    "--schema-include-matchable", is_flag=True, default=False, show_default=True
)
@click.argument("properties", nargs=-1)
def q(
    input_uri: str | None = "-",
    output_uri: str | None = "-",
    dataset: tuple[str] | None = (),
    schema: tuple[str] | None = (),
    schema_include_descendants: bool | None = False,
    schema_include_matchable: bool | None = False,
    properties: tuple[str] | None = (),
):
    """
    Apply ftmq filter to a json stream of ftm entities.
    """
    q = Query()
    for value in dataset:
        q = q.where(dataset=value)
    for value in schema:
        q = q.where(
            schema=value,
            include_descendants=schema_include_descendants,
            include_matchable=schema_include_matchable,
        )
    for prop, value, op in parse_unknown_cli_filters(properties):
        q = q.where(prop=prop, value=value, operator=op)

    proxies = q.apply_iter(smart_read_proxies(input_uri))
    smart_write_proxies(output_uri, proxies, serialize=True)


@cli.command("apply")
@click.option(
    "-i", "--input-uri", default="-", show_default=True, help="input file or uri"
)
@click.option(
    "-o", "--output-uri", default="-", show_default=True, help="output file or uri"
)
@click.option("-d", "--dataset", multiple=True, help="Dataset(s) to filter for")
@click.option("--replace-dataset", is_flag=True, default=False, show_default=True)
def apply(
    input_uri: str | None = "-",
    output_uri: str | None = "-",
    dataset: tuple[str] | None = (),
    replace_dataset: bool | None = False,
):
    """
    Uplevel an entity stream to nomenklatura entities and apply dataset(s) property
    """

    proxies = smart_read_proxies(input_uri)
    if dataset:
        proxies = apply_datasets(proxies, *dataset, replace=replace_dataset)
    smart_write_proxies(output_uri, proxies, serialize=True)
